----------------------------------------
-- Extremely Simple Text-mode Dialogs --
----------------------------------------
	
include graphics.e -- color names
include image.e -- save text image
include get.e
include wildcard.e

include keys.e

without warning

constant
	COLORS = 1,  FGND = 1, BKGND = 2,   -- box structure
	COORDS = 2,  ROW = 1, COL = 2,
	SIZE   = 3,  HEIGHT = 1, WIDTH = 2, MINH = 4, MINW = 8,
	STYLE  = 4,
	SHADOW = 5,
	HINT   = 6,
	HINTCOLORS = 7,
	BUFFER = 8,
	True = (1=1),
	False = not True
	
global constant
	SGL = 1, DBL = 2, DHL = 3, DVL = 4, NONE = 5, -- border styles
	CHECKBOX = 1, LISTBOX = 2, NOTIFY = 3, MSGBOX = 4, SELECT = 5 -- dialog types

object void, junk

sequence DefaultBox
	 DefaultBox =
	 	{
	 	 {BRIGHT_WHITE,BLUE},  -- default colors
		 {1,1},                -- coordinates top left corner
		 {5,10},               -- box size
		 SGL,                 -- single line border
		 1,                   -- light shadow
		 "Press any key...",  -- hint text
		 {BLACK,GRAY},       -- hint colors
		 {}
		}                  -- buffer for save screen

atom PX, PY, USER_ROUTINE
	PX = 1
	PY = 1
	
sequence dialogs, -- for tracking on-screen dialog boxes
		 bk_screen,
		 vc
		
	dialogs = {}
	bk_screen = {}

	vc = video_config()

function do_nothing()
	junk = message_box("Did Nothing","Doing Nothing",MB_OK)
	return {}
end function
USER_ROUTINE = routine_id( "do_nothing" )
		
global procedure printat(sequence coords, sequence text)
	position(coords[1],coords[2])
	puts(1,text)
end procedure

global procedure printcolor(sequence coords, sequence colors, sequence text)
	text_color(colors[1])
	bk_color(colors[2])
	printat(coords,text)
end procedure

global function limit(atom n, sequence limits)
	if n < limits[1] then
		n = limits[1]
	elsif n > limits[2] then
		n = limits[2]
	end if
	return n
end function

global function pad(sequence s, atom width)
	while length(s) < width do s = s & ' '
	end while
	return s
end function

global function delete(sequence s, atom loc, atom len)
	return s[1..loc-1] & s[loc+len..length(s)]
end function

procedure MoveCursor(sequence coords)
	position(coords[ROW],coords[COL])
end procedure

function SaveBkgnd(object box)
	cursor(NO_CURSOR)
	box[COORDS] = get_position()
	box[BUFFER] = save_text_image(box[COORDS],box[COORDS]+box[SIZE]+2)
	cursor(UNDERLINE_CURSOR)
	return box
end function

global procedure RestoreBkgnd(object box)
	MoveCursor(box[COORDS])
	display_text_image(box[COORDS],box[BUFFER])
end procedure

procedure SaveBackground()
	bk_screen = append(bk_screen, save_text_image({1,1},{25,80}) ) -- need screen size!
end procedure

procedure RestoreBackground()
	if length(bk_screen) > 0 then
		display_text_image( {1,1}, bk_screen[length(bk_screen)] )
		bk_screen = bk_screen[1..length(bk_screen)-1]
	end if
end procedure

-------------------------------< BOX >-----------------------------------
	
constant Lines = {{'','','','','',''}, -- single line borders
		  {'','','','','',''}, -- double line borders
		  {'','','','','',''}, -- double horizontal
		  {'','','','','',''}, -- double vertical
		  {' ',' ',' ',' ',' ',' '}} -- none
		
constant Divider = {{'','',''}, -- single style divider line
		    {'','',''},
		    {'','',''}, -- double horizontal
		    {'','',''},
		    {' ',' ',' '}}

constant Shadow = {'','',''}  -- 3 different shadow backgrounds
		
constant TL = 1, HL = 2, TR = 3, VL = 4, BL = 5, BR = 6

global function DisplayBox(sequence box)
sequence topline, botline, midline, divline, rshad, bshad, char
atom row, col, x, y
	box = SaveBkgnd(box)
	row = box[COORDS][ROW] col = box[COORDS][COL]
	x = box[SIZE][WIDTH] y = box[SIZE][HEIGHT]
	text_color(box[COLORS][FGND])
	bk_color(box[COLORS][BKGND])
	cursor(NO_CURSOR)

-- grab the area to be shadowed:
	rshad = save_text_image({row+1,col+x+2},{row+y+1,col+x+2})
	bshad = save_text_image({row+y+1,col+1},{row+y+1,col+x+1})

-- select the border style to use:
	char = Lines[box[STYLE]]

-- build the borders:
	topline = char[TL] & repeat(char[HL],x) & char[TR]
	botline = char[BL] & repeat(char[HL],x) & char[BR]
	midline = char[VL] & repeat(' ',x) & char[VL]
	char = Divider[box[STYLE]]
	divline = char[1] & repeat(char[2],x) & char[3]

-- draw the borders:
	printat({row,col}, topline)
	for i = 1 to y do
		printat({row+i,col}, midline)
	end for
	printat({row + 2,col}, divline)
	printat({row + y,col}, botline)

-- draw the shadows:
	if box[SHADOW] > 0 then
		position(20,1)
		for i = 1 to length(rshad) do
			rshad[i][2] = rshad[i][2] / 16
			rshad[i][1] = Shadow[box[SHADOW]]
		end for
		for i = 2 to length(bshad[1]) by 2 do
			bshad[1][i] = bshad[1][i] / 16
			bshad[1][i-1] = Shadow[box[SHADOW]]
		end for
		
		display_text_image({row+1,col+x+2},rshad)
		display_text_image({row+y+1,col+1},bshad)
	end if

	MoveCursor(box[COORDS]) -- restore cursor to start
	cursor(UNDERLINE_CURSOR)
	return box
end function

------------------------------------------------------------
global procedure SetColors(atom fgnd, atom bkgnd)
	DefaultBox[COLORS] = {fgnd,bkgnd}
end procedure

global procedure SetSize(atom rows, atom cols)
	DefaultBox[SIZE] = {rows,cols}
end procedure

global procedure SetStyle(atom style)
	DefaultBox[STYLE] = style
end procedure

global procedure SetHint(sequence hint)
	DefaultBox[HINT] = hint
end procedure

procedure DisplayHint(sequence hint)
	printcolor({25,2},DefaultBox[HINTCOLORS],repeat(' ',78))
	printcolor({25,3},DefaultBox[HINTCOLORS],hint)
end procedure

procedure DisplayList(sequence coords, sequence colors, sequence list, atom default)
	for i = 1 to length(list) do
		if i = default then
			printcolor(coords+{i,0},{colors[2],colors[1]},list[i])
		else
			printcolor(coords+{i,0},colors,list[i])
		end if
	end for
end procedure

function GenericBox(sequence title, sequence msg)
object msgbox
	msgbox = DefaultBox -- copy defaults
	msgbox = DisplayBox(msgbox)
	printat(msgbox[COORDS]+{1,1},title)
	DisplayList(msgbox[COORDS]+{2,1},msgbox[COLORS],msg,0)
	return msgbox
end function

global procedure MessageBox(sequence title, sequence msg)
object box
	box = GenericBox(title,msg)
end procedure

global procedure Notify(sequence title, sequence msg)
object box
	SaveBackground()
	box = GenericBox(title,msg)
	DisplayHint( "Press any key...")
	void = wait_key() -- wait for keypress
	RestoreBackground()
end procedure

global procedure set_user_routine( atom rid )
	USER_ROUTINE = rid
end procedure

global function CheckBox(sequence title, sequence items) -- revised by cklester
object box, selecteditems
atom key, selection, do_items, max, count_down, start_time, startitem, enditem,
	 redraw, counter, stopnow, spot, maxw, old_PY
sequence filter_buffer, ditem, sitem, flist

	SaveBackground()
	
	filter_buffer = ""
	count_down = 1
	start_time = time()
	
	box = GenericBox(title,"")
	max = box[SIZE][1] - 3
	maxw = box[SIZE][2]
	
	DisplayHint("Up/Down arrows to Select - Spacebar to toggle - ENTER when done")
	key = -1
	
	for i = 1 to length(items) do
		items[i] = "[ ] "& items[i]
	end for

	startitem = 1
	enditem = max
	if enditem > length(items) then
		enditem = length(items)
	end if
	selection = 1

	for i = startitem to enditem do
		printcolor(box[COORDS]+{i+2,1},box[COLORS],items[i] & repeat(' ',maxw-length(items[i]) ) )
	end for

	printcolor(box[COORDS]+{selection-startitem+1+2,1},box[COLORS],items[selection])
	MoveCursor(box[COORDS]+{selection-startitem+1+2,2})

	while key != ENTER and key != ESC do
		redraw = False
		key = get_key()
		if key = SPACE then
			if items[selection][2] = 'X' then
			   items[selection][2] = ' '
			else
			  items[selection][2] = 'X'
			end if
			redraw = True
		elsif key = F2 then -- run a user-defined item chooser
			old_PY = PY
			junk = call_func(USER_ROUTINE,{})
			if length(junk) > 0 then
				-- clear all first
				junk = upper(junk)
				for a=1 to length(items) do
					if items[a][2] = 'X' then
					   items[a][2] = ' '
					end if
				end for
				for t=1 to length(junk) do
					spot = find("[ ] " & junk[t], items)
					if spot > 0 then
						items[spot][2] = 'X'
					end if
				end for
			end if
			PY = old_PY
			redraw = True
			cursor(UNDERLINE_CURSOR)
		elsif key = HOME then
			selection = 1
			startitem = 1
			enditem = max
			redraw = True
		elsif key = KEND then
			selection = length(items)
			startitem = selection - max + 1
			enditem = length(items)
			redraw = True
		elsif key = UP then
			selection -= 1
			if selection < 1 then
				selection = 1
			else
				if selection < startitem then
					startitem -= 1
					enditem -= 1
				end if
				redraw = True
			end if
		elsif key = PGUP then
			selection -= max
			if selection < 1 then
				selection = 1
			end if
			startitem -= max
			enditem -=max
			if startitem < 1 then
				startitem = 1
				enditem = startitem + max - 1
			end if
			redraw = True
			
		elsif key = PGDN then
			selection += max
			if selection > length(items) then
				selection = length(items)
			end if
			enditem += max
			startitem += max
			if enditem > length(items) then
				enditem = length(items)
				startitem = enditem - max + 1
			end if
			redraw = True
		
		elsif key = DOWN then
			selection += 1
			if selection > length(items) then
				selection = length(items)
			else
				if selection > enditem then
					enditem += 1
					startitem += 1
				end if
				redraw = True
			end if
		end if
		key = upper(key)
		if find(key,"0123456789_-.ABCDEFGHIJKLMNOPQRSTUVWXYZ") > 0 then
			start_time = time() + count_down
			filter_buffer &= key
			counter = 1
			stopnow = False
			while not stopnow do
				if length(filter_buffer) + 4 <= length(items[counter]) then
					if equal(items[counter][5..4+length(filter_buffer)], filter_buffer) then
						selection = counter
						if counter + max > length(items) then
							counter = length(items) - max + 1
						end if
						startitem = counter
						enditem = startitem + max - 1
						stopnow = True
					end if
				end if
				counter += 1
				if counter = length(items) then stopnow = True end if
			end while
			redraw = True
		else
			if start_time < time() and length(filter_buffer) > 0 then
				filter_buffer = ""
				redraw = True
			end if
		end if
		
		if redraw then
			if length(items[selection]) > PY then
				sitem = items[selection][1..PY-1]
			else
				sitem = items[selection]
			end if
			
			for i = startitem to enditem do
				if length(items[i]) > PY then
					ditem = items[i][1..PY-1]
				else
					ditem = items[i]
				end if
				printcolor(box[COORDS]+{i+1-startitem+2,1},box[COLORS],ditem & repeat(' ',PY-length(ditem)) )
			end for

			printcolor(box[COORDS]+{selection-startitem+1+2,1},box[COLORS],sitem)
			MoveCursor(box[COORDS]+{selection-startitem+1+2,2})
		end if
	end while

	RestoreBackground()

	selecteditems = {}
	for i = 1 to length(items) do
		if items[i][2] = 'X' then
			selecteditems = append(selecteditems,items[i][5..length(items[i])])
		end if
	end for
	return selecteditems
end function

global function RadioBox(sequence title, sequence items)
object box, selecteditems
atom key, selection
 for i = 1 to length(items) do
     items[i] = "( ) "&items[i]
 end for
 box = GenericBox(title,items)
 DisplayHint("Select with up/dn arrows, toggle with spacebar, enter when finished")
 key = -1  selection = 1
 MoveCursor(box[COORDS]+{selection+2,2})
 while key != 13 do
       key = get_key()
       if key != -1 then
	  MoveCursor(box[COORDS]+{selection+2,2})
       end if
       if key = SPACE then
	  if items[selection][2] = 'X' then
	    items[selection][2] = ' '
	  else items[selection][2] = 'X'
	    printcolor(box[COORDS]+{selection+2,1},box[COLORS],items[selection])
	  end if
	  for i = 1 to length(items) do
	    if i != selection then
	     items[i][2] = ' '-- reset all buttons
	     printcolor(box[COORDS]+{i+2,1},box[COLORS],items[i])
	     end if
	    end for
	  printat(box[COORDS]+{selection+2,1},items[selection])
	  MoveCursor(box[COORDS]+{selection+2,2})
       elsif key = UP then
	  selection = limit(selection - 1,{1,length(items)})
	  MoveCursor(box[COORDS]+{selection+2,2})
       elsif key = DOWN then
	  selection = limit(selection + 1,{1,length(items)})
	  MoveCursor(box[COORDS]+{selection+2,2})
       elsif key = ESC then
	  exit
       end if
 end while
 selecteditems = {}
 for i = 1 to length(items) do
     if items[i][2] = 'X' then
	selecteditems = append(selecteditems,items[i][5..length(items[i])])
     end if
 end for
 return selecteditems
end function

global function ListBox(sequence title, sequence items, atom start) -- revised by cklester
object box
atom key, selection, do_items, max, count_down, start_time, startitem, enditem, redraw, counter, stopnow
sequence filter_buffer, ditem, sitem, flist

	SaveBackground()
		
	filter_buffer = ""
	count_down = 1
	start_time = time()
	
	box = GenericBox(title,"")
	max = box[SIZE][1] - 3

	DisplayHint("Up/Down arrows to Select - ENTER to Choose")
	key = -1
	
	startitem = 1
	enditem = max
	selection = 1
	if enditem > length(items) then
		enditem = length(items)
	end if
	for i = startitem to enditem do
		if length(items[i]) > PY then
			ditem = items[i][1..PY]
		else
			ditem = items[i] & repeat(' ',PY - length(items[i]))
		end if
		if i = selection then
			sitem = ditem
		end if
		printcolor(box[COORDS]+{i+1-startitem+2,1},box[COLORS],ditem )
	end for

	printcolor(box[COORDS]+{selection-startitem+1+2,1},{box[COLORS][2],box[COLORS][1]},sitem)

	cursor(NO_CURSOR)

	while key != ENTER and key != ESC do
		redraw = False
		key = get_key()
		if key = HOME then
			selection = 1
			startitem = 1
			enditem = max
			redraw = True
		elsif key = KEND then
			selection = length(items)
			startitem = selection - max + 1
			enditem = length(items)
			redraw = True
		elsif key = UP then
			selection -= 1
			if selection < 1 then
				selection = 1
			else
				if selection < startitem then
					startitem -= 1
					enditem -= 1
				end if
				redraw = True
			end if
		elsif key = PGUP then
			selection -= max
			if selection < 1 then
				selection = 1
			end if
			startitem -= max
			enditem -=max
			if startitem < 1 then
				startitem = 1
				enditem = startitem + max - 1
			end if
			redraw = True
			
		elsif key = PGDN then
			selection += max
			if selection > length(items) then
				selection = length(items)
			end if
			enditem += max
			startitem += max
			if enditem > length(items) then
				enditem = length(items)
				startitem = enditem - max + 1
			end if
			redraw = True
		
		elsif key = DOWN then
			selection += 1
			if selection > length(items) then
				selection = length(items)
			else
				if selection > enditem then
					enditem += 1
					startitem += 1
				end if
				redraw = True
			end if
		end if
		key = upper(key)
		if find(key,"0123456789_.ABCDEFGHIJKLMNOPQRSTUVWXYZ") > 0 then
			start_time = time() + count_down
			filter_buffer &= key
			counter = 1
			stopnow = False
			
			while not stopnow do
				if equal(upper(items[counter][1..length(filter_buffer)]), upper(filter_buffer)) then
					selection = counter
					if counter + max > length(items) then
						counter = length(items) - max + 1
					end if
					startitem = counter
					enditem = startitem + max - 1
					stopnow = True
				end if
				counter += 1
				if counter > length(items) then stopnow = True end if
			end while
			redraw = True
		else
			if start_time < time() and length(filter_buffer) > 0 then
				filter_buffer = ""
				redraw = True
			end if
		end if
		
		if redraw then
			for i = startitem to enditem do
				if length(items[i]) > PY then
					ditem = items[i][1..PY]
				else
					ditem = items[i] & repeat(' ',PY-length(items[i]))
				end if
				if i = selection then
					sitem = ditem
				end if
				printcolor(box[COORDS]+{i+1-startitem+2,1},box[COLORS],ditem )
			end for

			printcolor(box[COORDS]+{selection-startitem+1+2,1},{box[COLORS][2],box[COLORS][1]},sitem)
		end if
	end while

	RestoreBackground()

	return items[selection]
end function

global function Select(sequence title, sequence items, atom start) -- revised by cklester
object box
atom key, selection, do_items, max, count_down, start_time, startitem, enditem, redraw, counter, stopnow
sequence filter_buffer, ditem, sitem, flist

	SaveBackground()
		
	filter_buffer = ""
	count_down = 1
	start_time = time()
	
	box = GenericBox(title,"")
	max = box[SIZE][1] - 3

	DisplayHint("Up/Down arrows to Select - ENTER to Choose")
	key = -1
	
	startitem = 1
	enditem = max
	selection = 1
	if enditem > length(items) then
		enditem = length(items)
	end if
	for i = startitem to enditem do
		if length(items[i]) > PY then
			ditem = items[i][1..PY]
		else
			ditem = items[i] & repeat(' ',PY - length(items[i]))
		end if
		if i = selection then
			sitem = ditem
		end if
		printcolor(box[COORDS]+{i+1-startitem+2,1},box[COLORS],ditem )
	end for

	printcolor(box[COORDS]+{selection-startitem+1+2,1},{box[COLORS][2],box[COLORS][1]},sitem)

	cursor(NO_CURSOR)

	while key != ENTER and key != ESC do
		redraw = False
		key = get_key()
		if key = HOME then
			selection = 1
			startitem = 1
			enditem = max
			redraw = True
		elsif key = KEND then
			selection = length(items)
			startitem = selection - max + 1
			enditem = length(items)
			redraw = True
		elsif key = UP then
			selection -= 1
			if selection < 1 then
				selection = 1
			else
				if selection < startitem then
					startitem -= 1
					enditem -= 1
				end if
				redraw = True
			end if
		elsif key = PGUP then
			selection -= max
			if selection < 1 then
				selection = 1
			end if
			startitem -= max
			enditem -=max
			if startitem < 1 then
				startitem = 1
				enditem = startitem + max - 1
			end if
			redraw = True
			
		elsif key = PGDN then
			selection += max
			if selection > length(items) then
				selection = length(items)
			end if
			enditem += max
			startitem += max
			if enditem > length(items) then
				enditem = length(items)
				startitem = enditem - max + 1
			end if
			redraw = True
		
		elsif key = DOWN then
			selection += 1
			if selection > length(items) then
				selection = length(items)
			else
				if selection > enditem then
					enditem += 1
					startitem += 1
				end if
				redraw = True
			end if
		end if
		key = upper(key)
		if find(key,"0123456789_.ABCDEFGHIJKLMNOPQRSTUVWXYZ") > 0 then
			start_time = time() + count_down
			filter_buffer &= key
			counter = 1
			stopnow = False
			
			while not stopnow do
				if equal(upper(items[counter][1..length(filter_buffer)]), upper(filter_buffer)) then
					selection = counter
					if counter + max > length(items) then
						counter = length(items) - max + 1
					end if
					startitem = counter
					enditem = startitem + max - 1
					stopnow = True
				end if
				counter += 1
				if counter > length(items) then stopnow = True end if
			end while
			redraw = True
		else
			if start_time < time() and length(filter_buffer) > 0 then
				filter_buffer = ""
				redraw = True
			end if
		end if
		
		if redraw then
			for i = startitem to enditem do
				if length(items[i]) > PY then
					ditem = items[i][1..PY]
				else
					ditem = items[i] & repeat(' ',PY-length(items[i]))
				end if
				if i = selection then
					sitem = ditem
				end if
				printcolor(box[COORDS]+{i+1-startitem+2,1},box[COLORS],ditem )
			end for

			printcolor(box[COORDS]+{selection-startitem+1+2,1},{box[COLORS][2],box[COLORS][1]},sitem)
		end if
	end while

	RestoreBackground()

	return selection
end function

global function autoDialog(atom kind, sequence title, sequence items, atom start)
atom maxw, maxh, len, hi
	maxw = vc[VC_COLUMNS]
	maxh = vc[VC_LINES]-2
	len = length(title)
	for t=1 to length(items) do
		if length(items[t]) > len then
			len = length(items[t])
		end if
	end for
	if len > maxw then len = maxw end if
	hi = length(items) + 3
	if hi > maxh then hi = maxh end if
	SetSize( hi, len )
	PX = hi -- maximum box row
	PY = len -- max box column
	--center the dialog
	position( maxh/2 - hi/2 + 1, maxw/2 - len/2 )
	if kind = CHECKBOX then
		return CheckBox( title, items )
	elsif kind = LISTBOX then
		return ListBox( title, items, start )
	elsif kind = NOTIFY then
		Notify(title, items)
	elsif kind = MSGBOX then
		MessageBox(title,items)
	elsif kind = SELECT then
		return Select(title,items,start)
	end if
	return True
end function

function typeable(atom x)
	return not ( x < 31 or x > 126 )
end function

global function Input(sequence title, sequence prompt, sequence default)
sequence result, icoords, box
atom field_pos, filling,max
	max = prompt[2]
	prompt = prompt[1]
	box = GenericBox(title,prompt)
	icoords = get_position()
	printcolor(icoords,{box[COLORS][2],box[COLORS][1]},default)
	DisplayHint("Edit this field, press enter when done, or ESC to cancel")
	MoveCursor(icoords)
	result = default

	filling = True
	
	field_pos = length(default)

 	while filling do
		-- show field
		if field_pos = -1 then field_pos = 0 end if
		if field_pos > length(result) then field_pos = length(result) end if
		
		position(icoords[1],icoords[2])

		puts(1,result& repeat(' ',max-length(result)) )
		position(icoords[1],icoords[2]+field_pos)
		
		junk = wait_key()
		if junk = ESC then
			result = ""
			filling = (1=2)
		elsif junk = ENTER then
			filling = (1=2)
		elsif junk = 4 then -- CTRL+D
			result = ""
		elsif junk = LEFT then
			field_pos -= 1
		elsif junk = RIGHT then
			field_pos += 1
		elsif junk = DELETE then
			if length(result) > 0 and field_pos < length(result) then
				result = result[1..field_pos] & result[field_pos+2..length(result)]
			end if
		elsif junk = HOME then
			field_pos = 0
		elsif junk = KEND then
			field_pos = length(result)
		elsif junk = BACKSPACE then
			if length(result) > 0 and field_pos > 0 then
				result = result[1..field_pos-1] & result[field_pos+1..length(result)]
				field_pos -= 1
			end if
		elsif typeable( junk ) then
			if length(result) < max then
				result = result[1..field_pos] & junk & result[field_pos+1..length(result)]
				field_pos += 1
			end if
		end if
	end while
  RestoreBkgnd(box)
return result
end function

global procedure Progress(sequence title, sequence msg, atom percent)
object box, pbar
 box = GenericBox(title,{msg})
 pbar = repeat('',box[SIZE][COL] - 4)
 MoveCursor(box[COORDS]+{5,2})
 puts(1,pbar)
 pbar = repeat('',length(pbar) * percent * .01)
 MoveCursor(box[COORDS]+{5,2})
 puts(1,pbar)
end procedure

global procedure Delay(atom sec)
object start, fini
 start = time()
 fini = start + sec
 while time() < fini do
 end while
end procedure

global function list_show( sequence title, sequence list )
object choice
	position(3,5)
	SetSize(20,70) -- make this conform to screen size
	choice = ListBox(title, list , 1)
	bk_color(BLACK)
	text_color(WHITE)
	puts(1,"\n")
	return choice
end function
